/*
 Simple LCD driver. 8 bit interface, no R/W signal in use.

 OH2NLT / Juha Niinikoski 14.08.2002
 Modified for DC Receiver display & DDS board 11.01.2005

 S-meter logic (corrected) 12.01.2005

*/




/* 1x16 TM161ABA6 LCD display memory layout */
/* Organization is actually 2x8 characters in one display line */

#define LINE_1     0x00         /* line 1 start address, first 8 characters */
#define LINE_2     0x40         /* line 2 start address, next 8 characters */

#define CURSET     0x80         /* set memory cursor */
#define ADRSET	   0x40			/* set cgram address */

#define BUS_IDLE	0xFF		/* LCD Data Bus idle state */


/* print one character to LCD panel, RS = 1 */
void	lcdputch( unsigned char output )
	{
	LCD_DATA = output;
	LCD_RS = 1;
	LCD_E = 1;
	asm("nop");
	asm("nop");
	LCD_E = 0;
	LCD_DATA = BUS_IDLE;		/* leave bus idle state */
	delayu(50); 				/* about 40 us delay */
	}

/* send command to to LCD panel, RS = 0 */
void	lcdcom( char command )
	{
	LCD_DATA = command;
	LCD_RS = 0;
	LCD_E = 1;
	asm("nop");
	asm("nop");
	LCD_E = 0;
	LCD_RS = 1;					/* leave RS high */
	LCD_DATA = BUS_IDLE;		/* leave bus idle state */
	delayu(50); 				/* about 40 us delay */
	}

void lcdpos( char position )
/* set LCD cursor eg. LINE_1, LINE_2+5 */
	{
unsigned char a;
	a = CURSET + position;
	lcdcom( a );
	}

void lcdclr( void )
/* clear LCD panel & set cursor to 0.0 position */
	{
	lcdcom(0b00000001);	/* clear command */
	delaym(2);		/* long delay 2 mS */ 
	}

void clklcd( char output )
	{
	LCD_DATA = output;
	LCD_E = 1;
	asm("nop");
	asm("nop");
	LCD_E = 0;
	}

/* init LCD system */
void lcdinit( void )
	{
char i;
	delaym(20);

	for ( i = 0; i < 3; i++)
	{	
	clklcd(0x38);		/* LCD reset, 1-lines, 5*7 font  */
	delaym(6);
	}
	clklcd(0x08);		/* display off */
	delaym(5);
	clklcd(0x01);		/* display clear */
	delaym(5);
	clklcd(0x06);		/* entry mode set */
	delaym(5);

	clklcd(0x0C);		/* display on */
	delaym(5);
	}

/* TM161ABA6 character generator routines */

void set_cgram(unsigned char adr, unsigned char dta)	// set cgram address & data
	{	
	adr = adr & 0x3F;
	lcdcom(adr | ADRSET);					// set cgram address command + address
	lcdputch(dta);
	lcdclr();								// back to normal mode (clear LCD)	
	}


void set_ch_bits(char dta, char cnt)		// set chgen bits. Data, number of sets
	{
	char x;
	for(x=0; x<cnt; x++)
		{
		lcdputch(dta);
		}
	}	


void set_chgen(void)				// write bar symbols to HD44780 RAM chgen
	{		
	lcdcom(0 | ADRSET);				// set cgram address command + address

	set_ch_bits(0, 1);
	set_ch_bits(0x10, 5);			// CGRAM 00 Blocks L > R
	set_ch_bits(0, 2);

	set_ch_bits(0, 1);
	set_ch_bits(0x14, 5);			// CGRAM 01
	set_ch_bits(0, 2);

	set_ch_bits(0, 1);
	set_ch_bits(0x15, 5);			// CGRAM 02 "Full block"
	set_ch_bits(0, 2);

	set_ch_bits(0, 1);
	set_ch_bits(0x01, 5);			// CGRAM 03 Blocks R > L
	set_ch_bits(0, 2);

	set_ch_bits(0, 1);
	set_ch_bits(0x05, 5);			// CGRAM 04
	set_ch_bits(0, 2);
							
	set_ch_bits(0x08, 7);			// CGRAM 05 spare
	set_ch_bits(0, 1);

	set_ch_bits(0, 3);				// CGRAM 06 S-meter scale dot
	set_ch_bits(0x10, 1);
	set_ch_bits(0, 4);

	set_ch_bits(0, 2);				// CGRAM 07 S-meter scale bar
	set_ch_bits(0x10, 3);
	set_ch_bits(0, 3);

	lcdclr();						// back to normal mode (clear LCD)
	}


// Meter constants

#define DISP_LEN 8								// display lemgth (char)
#define FONT_W 6								// font width, corrected 12.01.2005
#define MAX_GRAPH (DISP_LEN * FONT_W)			// graph length in pixels (5*8 = 48)

const char scale_dots[8] = {6,32,6,32,7,32,7,32};				// scale dots for S-meter

const char Blockslr[7] = {0x20, 0x00, 0x00, 0x01, 0x01, 0x02, 0x02};	// blocks for L to R bar graph / 12.01.2005

// draw S-meter, scale = 0...48
// buffer is used to avoid LCD flicker

void draw_s_meter(char len)
	{
	char x;
	char whole;
	char part;
	char idx;
	char bar_buffer[DISP_LEN];				// display buffer
	
	if(len > MAX_GRAPH)						// cut to max
		len = MAX_GRAPH;

	whole = len / FONT_W;					// calculate graph
	part = len % FONT_W;

	idx = 0;
	for(x=0; x<DISP_LEN; x++)				// copy scale dots to buffer
		{
		bar_buffer[idx] = scale_dots[x];
		idx++;
		}

	idx = 0;								// draw graph to buffer	
	while(whole > 0)			
		{
		bar_buffer[idx] = 0x02;				// draw full blocks
		idx++;	
		whole--;
		}

	if(part > 0)
		bar_buffer[idx] = Blockslr[part]; 	// draw partial block to buffer

	for(x=0; x<DISP_LEN; x++)				// print meter buffer to LCD
		{
		lcdputch(bar_buffer[x]);
		}
	}



/* put string  */

void putst(register const char *str)
{
	while((*str)!=0)
	{
		lcdputch(*str);
		str++;
	}
}

/* write character in hex format */

void putchhex(unsigned char c)
{
	unsigned char temp;

	// transmits in hex

	temp=c;

	c=(c >> 4);
	if (c<10) c+=48; else c+=55;
	lcdputch(c);

	c=temp;

	c=(c & 0x0F);
	if (c<10) c+=48; else c+=55;
	lcdputch(c);
}

/* write long in hex format */

void putlonghex( long ww )
	{

	union
	{
	long w;
	unsigned char p[4];
	} lrek;

	lrek.w = ww;
	putchhex(lrek.p[3]);
	putchhex(lrek.p[2]);
	putchhex(lrek.p[1]);
	putchhex(lrek.p[0]);
	}


/* Special frequency display format format 3.699.00 */
/* 1Hz digit has left out, display overflow not handeled  */

void putfreq( long x)
	{

	//million
	lcdputch((x/1000000)+'0');		
	x-=(x/1000000)*1000000;

	lcdputch('.');

	//100_thousands
	lcdputch((x/100000)+'0');
	x-=(x/100000)*100000;
	
	//10_thousands
	lcdputch((x/10000)+'0');
	x-=(x/10000)*10000;

	//thousands
	lcdputch((x/1000)+'0');
	x-=(x/1000)*1000;

	lcdputch('.');

	//hundreds
	lcdputch((x/100)+'0');
	x-=(x/100)*100;

	//tens
	lcdputch((x/10)+'0');
	x-=(x/10)*10;

	//ones
//	putch((x/1)+'0');
	}


/* Special Voltage display format 13.80 */

void putvolt( int x)
	{
	//thousands
	lcdputch((x/1000)+'0');
	x-=(x/1000)*1000;

	//hundreds
	lcdputch((x/100)+'0');
	x-=(x/100)*100;

	lcdputch('.');

	//tens
	lcdputch((x/10)+'0');
	x-=(x/10)*10;

	//ones
	lcdputch(x+'0');
	}
