/*
 Simple LCD driver. 8 bit interface, no R/W signal in use.

 OH2NLT / Juha Niinikoski 14.08.2002
 Modified for DC Receiver display & DDS board 11.01.2005

 S-meter logic (corrected) 12.01.2005
 New code size optimized decimal number conversion. Code size reduced 450 words. 19.01.2005
 HEX nibble display added, code size optimized 04.02.2005
 s-meter code size optimized 07.02.2005
 small code size addjustment 13.02.2005

*/




/* 1x16 TM161ABA6 LCD display memory layout */
/* Organization is actually 2x8 characters in one display line */

#define LINE_1     0x00         /* line 1 start address, first 8 characters */
#define LINE_2     0x40         /* line 2 start address, next 8 characters */

#define CURSET     0x80         /* set memory cursor */
#define ADRSET	   0x40			/* set cgram address */

#define FIRST_HALF (CURSET + LINE_1)	/* first half of display */
#define SECOND_HALF (CURSET + LINE_2)	/* second half of display */

#define BUS_IDLE	0xFF		/* LCD Data Bus idle state */


/* print one character to LCD panel, RS = 1 */
void	lcdputch( unsigned char output )
	{
	LCD_DATA = output;
	LCD_RS = 1;
	LCD_E = 1;
	asm("nop");
	asm("nop");
	LCD_E = 0;
	LCD_DATA = BUS_IDLE;		/* leave bus idle state */
	delayu(50); 				/* about 40 us delay */
	}

/* send command to to LCD panel, RS = 0 */
void	lcdcom( char command )
	{
	LCD_DATA = command;
	LCD_RS = 0;
	LCD_E = 1;
	asm("nop");
	asm("nop");
	LCD_E = 0;
	LCD_RS = 1;					/* leave RS high */
	LCD_DATA = BUS_IDLE;		/* leave bus idle state */
	delayu(50); 				/* about 40 us delay */
	}

#if 0
void lcdpos( char position )	/* position cursor */
/* set LCD cursor eg. LINE_1, LINE_2+5 */
	{
unsigned char a;
	a = CURSET + position;
	lcdcom( a );
	}
#endif

void lcdclr( void )
/* clear LCD panel & set cursor to 0.0 position */
	{
	lcdcom(0b00000001);		/* clear command */
	delaym(2);				/* long delay 2 mS */ 
	}

void clklcd( char output )
	{
	LCD_DATA = output;
	LCD_E = 1;
	asm("nop");
	asm("nop");
	LCD_E = 0;
	}

/* init LCD system */
void lcdinit( void )
	{
char i;
	delaym(20);

	for ( i = 0; i < 3; i++)
	{	
	clklcd(0x38);		/* LCD reset, 1-lines, 5*7 font  */
	delaym(6);
	}
	clklcd(0x08);		/* display off */
	delaym(5);
	clklcd(0x01);		/* display clear */
	delaym(5);
	clklcd(0x06);		/* entry mode set */
	delaym(5);

	clklcd(0x0C);		/* display on */
	delaym(5);
	}

/* TM161ABA6 character generator routines */

void set_cgram(unsigned char adr, unsigned char dta)	// set cgram address & data
	{	
	adr = adr & 0x3F;
	lcdcom(adr | ADRSET);					// set cgram address command + address
	lcdputch(dta);
	lcdclr();								// back to normal mode (clear LCD)	
	}

void set_ch_bits(char dta, char cnt)		// set chgen bits. Data, number of sets
	{
	char x;
	for(x=0; x<cnt; x++)
		{
		lcdputch(dta);
		}
	}	

// row count must be 8 for each character

void set_chgen(void)				// write bar symbols to HD44780 RAM chgen
	{		
	lcdcom(0 | ADRSET);				// set cgram address command + address

	set_ch_bits(0, 1);				// Blocks L > R
	set_ch_bits(0x10, 5);			// CGRAM 00 "one bar block"
	set_ch_bits(0, 2);

	set_ch_bits(0, 1);
	set_ch_bits(0x14, 5);			// CGRAM 01 "two bar block"
	set_ch_bits(0, 2);

	set_ch_bits(0, 1);
	set_ch_bits(0x15, 5);			// CGRAM 02 "Full block"
	set_ch_bits(0, 2);

	set_ch_bits(0, 3);				// CGRAM 03 S-meter scale dot
	set_ch_bits(0x10, 1);
	set_ch_bits(0, 4);

	set_ch_bits(0, 2);				// CGRAM 04 S-meter scale bar
	set_ch_bits(0x10, 3);
	set_ch_bits(0, 3);

	lcdclr();						// back to normal mode (clear LCD)
	}


// S-meter constants

#define DISP_LEN 8									// display lemgth (char)
#define FONT_W 6									// font width, corrected 12.01.2005
#define MAX_GRAPH (DISP_LEN * FONT_W)				// graph length in pixels (5*8 = 48)

const char scale_dots[8] = {3,32,3,32,4,32,4,32};	// scale dots for S-meter

const char Blockslr[7] = {0x20, 0x00, 0x00, 0x01, 0x01, 0x02, 0x02};	// blocks for L to R bar graph / 12.01.2005

// draw S-meter, scale = 0...48
// buffer is used to avoid LCD flicker

void draw_s_meter(char len)
	{
	char x;
	char whole;
	char part;
	char idx;
	char bar_buffer[DISP_LEN];				// display buffer
	
	if(len > MAX_GRAPH)						// cut to max
		len = MAX_GRAPH;

	whole = len / FONT_W;					// calculate graph
	part = len % FONT_W;

	idx = 0;
	for(x=0; x<DISP_LEN; x++)				// copy scale dots to buffer
		{
		bar_buffer[idx] = scale_dots[x];
		idx++;
		}

	idx = 0;								// draw graph to buffer	
	while(whole > 0)			
		{
		bar_buffer[idx] = 0x02;				// draw full blocks
		idx++;	
		whole--;
		}

	if(part > 0)
		bar_buffer[idx] = Blockslr[part]; 	// draw partial block to buffer

	for(x=0; x<DISP_LEN; x++)				// print meter buffer to LCD
		{
		lcdputch(bar_buffer[x]);
		}
	}



/* put string  */

void putst(register const char *str)
{
	while((*str)!=0)
	{
		lcdputch(*str);
		str++;
	}
}


/* display 4-bit nibble in hex format */

void puthexnible( unsigned char c )
	{
	c=(c & 0x0F);					// mask high nibble out
	if (c<10) c+=48; else c+=55;	// HEX convert
	lcdputch(c);					// display
	}


/* display 8-bit character in hex format */

void putchhex(unsigned char c)
{
	unsigned char cl;

	cl=c;
	c=(c >> 4);						// shift high nibble down
	puthexnible(c);					// display high nibble
	puthexnible(cl);				// display low nibble
}


/* display long in hex format */

void putlonghex( long ww )
	{

	union
	{
	long w;
	unsigned char p[4];
	} lrek;

	lrek.w = ww;
	putchhex(lrek.p[3]);
	putchhex(lrek.p[2]);
	putchhex(lrek.p[1]);
	putchhex(lrek.p[0]);
	}


unsigned char disp_buffer[8];		// display buffer for long number printouts

/* Convert binary number to ASCII string */
/* Only 7 digits are converted */

void bin2ascii(unsigned long luku)
	{
char x, idx;

	idx = 6;
	for(x=0; x<7; x++)				// 7 digits, 0.000.000
		{
		disp_buffer[idx] = (luku % 10) + '0';
		luku = luku / 10;
		idx--;
		}
	disp_buffer[7] = 0;				// string terminator
	}


/* Special frequency display format format 3.699.00 */
/* 1Hz digit has left out, display overflow not handeled  */

void putfreq( long x)
	{
	bin2ascii(x);					// partial convert long number to string

	lcdputch(disp_buffer[0]);		// MHz
	lcdputch('.');
	lcdputch(disp_buffer[1]);		// 100kHz
	lcdputch(disp_buffer[2]);		// 10kHz
	lcdputch(disp_buffer[3]);		// 1kHz
	lcdputch('.');
	lcdputch(disp_buffer[4]);		// 100Hz
	lcdputch(disp_buffer[5]);		// 10Hz
									// 1Hz not displayed
	}


/* Special Voltage display format 13.80 */

void putvolt( int x)
	{
	bin2ascii(x);					// convert number to string

	lcdputch(disp_buffer[3]);		// 10V
	lcdputch(disp_buffer[4]);		// 1V
	lcdputch('.');
	lcdputch(disp_buffer[5]);		// 100mV
	lcdputch(disp_buffer[6]);		// 10mV
	}
